#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import numpy as np

from .autoplumber import AutoPlumber


class PretrainedModel:
    """A pretrained model, built using scivision configuration file."""

    def __init__(self, config: dict):
        self._config = config
        self._plumbing = AutoPlumber(config)

    def predict(self, X: np.ndarray, *args, **kwargs):
        """Run the prediction."""
        # note, we may need to pass other kwargs to the prediction here
        return self._plumbing(X, **kwargs)

    def __repr__(self):
        """Return a nice representation of the model."""
        return (
            f"scivision.PretrainedModel( \n"
            f"  module='{self._plumbing.module.__name__}', \n"
            f"  model='{self._plumbing.model.__class__.__name__}', \n"
            f"  source='{self._config['url']}' \n"
            f"  pipe='{self._plumbing.pipe}' \n)"
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pkgutil
import pandas as pd
import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, Optional, Tuple, Union
from pydantic import BaseModel, AnyUrl, FileUrl
from enum import Enum


class TaskEnum(str, Enum):
    classificiation = "classification"
    object_detection = "object-detection"
    segmentation = "segmentation"
    thresholding = "thresholding"
    other = "other"


class CatalogModelEntry(BaseModel, extra="forbid"):
    # tasks, institution and tags are Tuples (rather than Lists) so
    # that they are immutable - Tuple is being used as an immutable
    # sequence here. This means that these fields are hashable, which
    # can be more convenient when included in a dataframe
    # (e.g. unique()). Could consider using a Frozenset for these
    # fields instead, since duplicates and ordering should not be
    # significant.
    name: str
    description: Optional[str]
    tasks: Tuple[TaskEnum, ...]
    url: Union[AnyUrl, FileUrl]
    pkg_url: str  # may be a PyPI package name (not an AnyUrl)
    format: str
    pretrained: bool
    labels_required: bool
    institution: Tuple[str, ...] = ()
    tags: Tuple[str, ...]

    def __getitem__(self, item):
        return getattr(self, item)


class CatalogModels(BaseModel, extra="forbid"):
    catalog_type: str = "scivision model catalog"
    name: str
    # Tuple: see comment on CatalogModelEntry
    entries: Tuple[CatalogModelEntry, ...]


class CatalogDatasourceEntry(BaseModel, extra="forbid"):
    name: str
    description: Optional[str]
    # Tuple: see comment on CatalogModelEntry
    tasks: Tuple[TaskEnum, ...]
    domains: Tuple[str, ...]
    url: Union[AnyUrl, FileUrl]
    format: str
    labels_provided: bool
    institution: Tuple[str, ...] = ()
    tags: Tuple[str, ...]

    def __getitem__(self, item):
        return getattr(self, item)


class CatalogDatasources(BaseModel, extra="forbid"):
    catalog_type: str = "scivision datasource catalog"
    name: str
    # Tuple: see comment on CatalogModelEntry
    entries: Tuple[CatalogDatasourceEntry, ...]


def _coerce_datasources_catalog(
    datasources: Union[CatalogDatasources, os.PathLike, None]
) -> CatalogDatasources:
    """Returns a CatalogDatasources determined from the argument: either
    the one passed, or one loaded from a file
    """
    if isinstance(datasources, CatalogDatasources):
        return datasources
    elif isinstance(datasources, (bytes, str, os.PathLike)):
        datasources_raw = Path(datasources).read_text()
        return CatalogDatasources.parse_raw(datasources_raw)
    elif datasources is None:
        datasources_raw = pkgutil.get_data(__name__, "data/datasources.json")
        return CatalogDatasources.parse_raw(datasources_raw)
    else:
        raise TypeError("Cannot load datasource from unsupported type")


def _coerce_models_catalog(
    models: Union[CatalogModels, os.PathLike, None]
) -> CatalogModels:
    """Returns a CatalogModels determined from the argument: either the
    one passed, or one loaded from a file
    """
    if isinstance(models, CatalogModels):
        return models
    elif isinstance(models, (bytes, str, os.PathLike)):
        models_raw = Path(models).read_text()
        return CatalogModels.parse_raw(models_raw)
    elif models is None:
        models_raw = pkgutil.get_data(__name__, "data/models.json")
        return CatalogModels.parse_raw(models_raw)
    else:
        raise TypeError("Cannot load datasource from unsupported type")


class QueryResult(ABC):
    @abstractmethod
    def to_dataframe(self) -> pd.DataFrame:
        ...

    def to_dict(self) -> Dict[str, Any]:
        return self.to_dataframe().to_dict(orient="records")


class PandasQueryResult(QueryResult):
    def __init__(self, result: pd.DataFrame):
        self._result = result

    def to_dataframe(self) -> pd.DataFrame:
        return self._result


class PandasCatalog:
    def __init__(self, datasources=None, models=None):
        super().__init__()

        if isinstance(datasources, pd.DataFrame):
            self._datasources = datasources
        else:
            datasources_cat = _coerce_datasources_catalog(datasources)
            self._datasources = pd.DataFrame(
                [ent.dict() for ent in datasources_cat.entries]
            )

        if isinstance(models, pd.DataFrame):
            self._models = models
        else:
            models_cat = _coerce_models_catalog(models)
            self._models = pd.DataFrame([ent.dict() for ent in models_cat.entries])

    @property
    def models(self) -> PandasQueryResult:
        return PandasQueryResult(self._models)

    @property
    def datasources(self) -> PandasQueryResult:
        return PandasQueryResult(self._datasources)

    def _compatible_models(self, datasource) -> PandasQueryResult:
        models_compatible_format = self._models[
            self._models.format == datasource["format"]
        ]

        models_compatible_format_labels = models_compatible_format[
            datasource["labels_provided"] | ~models_compatible_format.labels_required
        ]

        datasource_tasks = pd.DataFrame(datasource["tasks"], columns=["tasks"])
        models_compatible_tasks = (
            self._models[["name", "tasks"]]
            .explode("tasks")
            .merge(datasource_tasks, on="tasks", suffixes=("_model", "_datasource"),)
            .name.drop_duplicates()
        )
        result_df = models_compatible_format_labels[
            models_compatible_format_labels.name.isin(models_compatible_tasks)
        ]

        return PandasQueryResult(result_df)

    # Similar to _compatible_models, but for datasources.  Can't
    # cleanly combine these two functions, due to the asymmetry
    # between a model's 'labels_required', and datasource 'labels_provided'.
    # In particular, a model that doesn't require labels can still use
    # a datasource that provides them (if they are otherwise
    # compatible), but not vice versa.  For this reason, the distinct
    # names 'labels_provided' and 'labels_required' are used.
    def _compatible_datasources(self, model) -> PandasQueryResult:
        datasources_compatible_format = self._datasources[
            self._datasources.format == model["format"]
        ]

        datasources_compatible_format_labels = datasources_compatible_format[
            datasources_compatible_format.labels_provided | ~model["labels_required"]
        ]

        model_tasks = pd.DataFrame(model["tasks"], columns=["tasks"])
        datasources_compatible_tasks = (
            self._datasources[["name", "tasks"]]
            .explode("tasks")
            .merge(model_tasks, on="tasks", suffixes=("_model", "_datasource"),)
            .name.drop_duplicates()
        )
        result_df = datasources_compatible_format_labels[
            datasources_compatible_format_labels.name.isin(datasources_compatible_tasks)
        ]

        return PandasQueryResult(result_df)

    def compatible_models(self, datasource) -> PandasQueryResult:
        """Return all models that are compatible with datasource

        Parameters
        ----------
        datasource : str or dict-like
            Any dictionary-like (including CatalogDatasourceEntry) that
            has keys 'format', 'tasks' and 'labels_provided', representing
            these properties of the datasource.
            If a string is passed, this is used to look up the datasource
            (in `self._datasources`).

        Returns
        -------
        result: QueryResult
            A QueryResult instance containing the models compatible with the
            given datasource (convertible to a dict or pd.DataFrame).

        """
        if isinstance(datasource, str):
            return self._compatible_models(
                self._datasources.set_index("name").loc[datasource]
            )
        else:
            return self._compatible_models(datasource)

    def compatible_datasources(self, model) -> PandasQueryResult:
        """Return all datasources that are compatible with model

        Parameters
        ----------
        model : str or dict-like
            Any dictionary-like (including CatalogModelEntry) that has
            keys 'format', 'tasks' and 'labels_required', representing
            these properties of the model.
            If a string is passed, this is used to look up the model (in `self._models`).

        Returns
        -------
        result: QueryResult
            A QueryResult instance containing the datasources compatible with
            the given model (convertible to a dict or pd.DataFrame).

        """
        if isinstance(model, str):
            return self._compatible_datasources(
                self._models.set_index("name").loc[model]
            )
        else:
            return self._compatible_datasources(model)


default_catalog = PandasCatalog()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import importlib
import inspect
from typing import NamedTuple

import numpy as np


class DataPipe(NamedTuple):
    input: inspect.Parameter
    output: inspect.Parameter


class AutoPlumber:
    """Automagically plumb a model using the scivision config file.

    This class will inspect the module and try to work out how to
    map the inputs of the Callable to the scivision function
    signature.

    Attributes
    ----------
    config : dict
        The scivision config as a dictionary.

    """

    def __init__(self, config: dict):

        # import the module and get the model object
        self._module = importlib.import_module(config["import"])
        model = getattr(self._module, config["model"])

        # we could instantiate the model using the args here
        # args = config["args"]
        self._model = model()

        self._fn = getattr(self._model, config["prediction_fn"]["call"])

        # get the call signature
        self._model_signature = inspect.signature(self._fn)
        model_input = config["prediction_fn"]["args"]["X"]
        try:
            model_param = self._model_signature.parameters[model_input]
        except KeyError:
            raise KeyError(
                f"Parameter `{model_input}` not found in "
                f"`{self._model.__class__.__name__}`. There was an error "
                f"parsing the `.scivision-config.yaml` configuration file."
            )

        # this is a bit weird as we're determining our own signature, but...
        X = inspect.signature(self).parameters["X"]
        self._pipe = DataPipe(X, model_param)

    @property
    def module(self):
        return self._module

    @property
    def model(self):
        return self._model

    @property
    def pipe(self):
        return self._pipe

    def __call__(self, X: np.ndarray, **kwargs):
        """Redirect the input X to the correct input of the model."""

        # TODO(arl): this is where the actual piping needs to happen
        model_args = []
        model_kwargs = {self.pipe.output.name: X}

        # optionally, we can also pass on other keyword arguments to the model
        model_kwargs.update(kwargs)
        return self._fn(*model_args, **model_kwargs)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import importlib
import subprocess
import sys


def _package_exists(config: dict) -> bool:
    """Check to see whether a package exists."""
    try:
        importlib.import_module(config["import"])
    except ModuleNotFoundError:
        return False

    return True


def package_from_config(config: dict, branch: str = "main") -> str:
    """Given a config return the pip install string."""
    install_str = config["url"]
    if install_str.endswith(".git"):
        install_str = install_str[:-4]
    install_branch = config.get("github_branch", branch)
    return f"git+{install_str}@{install_branch}#egg={config['import']}"


def _install(package):
    """Install a package using pip."""
    subprocess.check_call([sys.executable, "-m", "pip", "install", package])


def _reinstall(package):
    """Reinstall a package using pip."""
    subprocess.check_call([sys.executable, "-m", "pip", "install",
                           "--force-reinstall", "--no-deps", package])


def install_package(config: dict, allow_install: bool = False, branch: str = "main"):
    """Install the python package if it doesn't exist."""
    package = package_from_config(config, branch)
    if _package_exists(config):  # only reinstall if allow_install is True
        if allow_install:
            _reinstall(package)
    else:
        if allow_install:
            _install(package)
        else:
            raise Exception(
                "Package does not exist. Try installing it with: \n"
                f"`!pip install -e {package}`"
            )



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import inspect
import io
import logging
from functools import wraps
from typing import Callable


KOALA_LOGGING_LEVEL = logging.DEBUG

# get the logger instance
logger = logging.getLogger(__name__)
stream = io.StringIO()


# if we don't have any handlers, set one up
if not logger.handlers:
    format = logging.Formatter(
        "\U0001f428 [%(levelname)s][%(asctime)s] %(message)s",
        datefmt="%Y/%m/%d %H:%M:%S",
    )
    handler = logging.StreamHandler(stream)
    handler.setFormatter(format)
    logger.addHandler(handler)
    logger.setLevel(KOALA_LOGGING_LEVEL)


def koala(fn: Callable) -> Callable:
    """Decorator to enable logging of calls to core `scivision` functions.

    Parameters
    ----------
    fn : Callable
        The function to be wrapped with the `koala` logger.

    Returns
    -------
    wrapped_fn : Callable
        The `koala` wrapped function.
    """

    @wraps(fn)
    def wrapped_fn(*args, **kwargs):

        # we can also log the signature and kwargs provided
        signature = inspect.signature(fn)
        logger.info(
            f"Call='{fn.__name__}', module='{inspect.getmodule(fn).__name__}'"
            f", signature={signature}"
            f", kwargs_given='{', '.join(kwargs.keys())}'"
        )

        # now call the function, but log any exceptions
        try:
            output = fn(*args, **kwargs)
        except Exception as e:
            logger.error(e)
            raise
        return output

    return wrapped_fn


def get_koala_log() -> str:
    """Get the koala log.

    Returns
    -------
    stream : str
        The koala log stream as a string.
    """
    return stream.getvalue()



#!/usr/bin/env python
from setuptools import find_packages, setup

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

requirements = []
with open("requirements.txt") as f:
    for line in f:
        stripped = line.split("#")[0].strip()
        if len(stripped) > 0:
            requirements.append(stripped)

setup(
    name="scivision",
    version="0.2.5",
    description="scivision",
    author="The Alan Turing Institute",
    author_email="scivision@turing.ac.uk",
    url="https://github.com/alan-turing-institute/scivision",
    packages=find_packages(),
    install_requires=requirements,
    python_requires=">=3.7",
    include_package_data=True,
    long_description=long_description,
    long_description_content_type='text/markdown',
)
